# bloom-filter


## Описание

Фильтр Блума - это структура данных, которая позволяет проверить принадлежность элемента множеству. Особенностью её работы является то, что возможны ложноположительные (подтверждение принадлежности, когда её нет), но не ложноотрицательные (отрицание принадлежности, когда она есть) срабатывания.

Мы не будем углубляться в особенности работы структуры, а используем её в качестве разминки перед более сложными задачами.


## Задание

В файле `public/bloom_filter.h` можно найти саму структуру `BloomFilter` и объявление связанных с ней фунций `calc_hash`, `bloom_init`, `bloom_destroy`, `bloom_insert` и `bloom_check`.


Предлагается реализация со следующими особенностями:

- Элементами множества являются С-строки
- Множество взведённых битов хранится в массиве 64-битных чисел `BloomFilter.set`
- Размер множества равен `BloomFilter.size` и может быть произвольным
- Для простоты мы реализуем только одну хэш-функцию для строк (например, полиномиальную), которая возвращает значения из отрезка `[0 .. modulus-1]`, при этом у вас есть возможность параметризовать её с помощью аргумента `seed` (это пригодится, поскольку `BloomFilter.hash_fn_count`, число хэш-функций, не обязательно равно единице)


От вас требуется реализовать функции `calc_hash`, `bloom_init`, `bloom_destroy`, `bloom_insert` и `bloom_check`. Шаблон решения находится в файле `template/bloom_filter.c`.


## Тесты

Протестировать своё решение можно с помощью публичных тестов, описанных в `public/tester.c`. Они достаточно простые, но помогут убедиться в том, что все функции реализованы правильно.


## Дополнительная информация

- [Статья на Википедии](https://en.wikipedia.org/wiki/Bloom_filter)
- [Bigtable: A Distributed Storage System for Structured Data](https://research.google/pubs/pub27898/) - применение фильтра Блума в БД BigTable для уменьшения количества обращений к диску
- [Language Models are Unsupervised Multitask Learners](https://cdn.openai.com/better-language-models/language_models_are_unsupervised_multitask_learners.pdf) - применение фильтра Блума при обучении GPT-2 для оценки утечки тестовых данных в обучающую выборку
